"""doc
# pytorch_mjolnir.Experiment / mjolnir_experiment

> A lightning module that runs an experiment in a managed way.

There is first the Experiment base class from wich all experiments must inherit (directly or indirectly).
"""
import os
from typing import Any, Tuple
import pytorch_lightning as pl
import torch
import argparse
from torch.utils.data.dataloader import DataLoader
from time import time
from datetime import datetime
from pytorch_lightning.loggers.tensorboard import TensorBoardLogger
from torch.utils.data.dataset import IterableDataset


def _generate_version() -> str:
    return datetime.fromtimestamp(time()).strftime('%Y-%m-%d_%H.%M.%S')


def run(experiment_class):
    """
    You can use this main function to make your experiment runnable with command line arguments.

    Simply add this to the end of your experiment.py file:

    ```python
    if __name__ == "__main__":
        from pytorch_mjolnir import run
        run(MyExperiment)
    ```

    Then you can call your python file from the command line and use the help to figure out the parameters.
    ```bash
    python my_experiment.py --help
    ```
    """
    gpus = 1
    if "SLURM_GPUS" in os.environ:
        gpus = int(os.environ["SLURM_GPUS"])
    nodes = 1
    if "SLURM_NODES" in os.environ:
        nodes = int(os.environ["SLURM_NODES"])
    output_path = "logs"
    if "RESULTS_PATH" in os.environ:
        output_path = os.environ["RESULTS_PATH"]
    
    parser = argparse.ArgumentParser(description='The main entry point for the script.')
    parser.add_argument('--name', type=str, required=True, help='The name for the experiment.')
    parser.add_argument('--version', type=str, required=False, default=None, help='The version that should be used (defaults to timestamp).')
    parser.add_argument('--output', type=str, required=False, default=output_path, help='The name for the experiment (defaults to $RESULTS_PATH or "logs").')
    parser.add_argument('--gpus', type=int, required=False, default=gpus, help='Number of GPUs that can be used.')
    parser.add_argument('--nodes', type=int, required=False, default=nodes, help='Number of nodes that can be used.')
    parser.add_argument('--resume_checkpoint', type=str, required=False, default=None, help='A specific checkpoint to load. If not provided it tries to load latest if any exists.')
    args, other_args = parser.parse_known_args()

    experiment = experiment_class(**other_args)
    experiment.run_experiment(name=args.name, version=args.version, output_path=args.output, resume_checkpoint=args.resume_checkpoint, gpus=args.gpus, nodes=args.nodes)


class Experiment(pl.LightningModule):
    """
    An experiment base class.

    All experiments must inherit from this.
    
    ```python
    from pytorch_mjolnir import Experiment
    class MyExperiment(Experiment):
        [...]
    ```
    """
    def run_experiment(self, name: str, gpus: int, nodes: int, version=None, output_path=os.getcwd(), resume_checkpoint=None):
        """
        Run the experiment.

        :param name: The name of the family of experiments you are conducting.
        :param gpus: The number of gpus used for training.
        :param nodes: The number of nodes used for training.
        :param version: The name for the specific run of the experiment in the family (defaults to a timestamp).
        :param output_path: The path where to store the outputs of the experiment (defaults to the current working directory).
        :param resume_checkpoint: The path to the checkpoint that should be resumed (defaults to None).
            In case of None this searches for a checkpoint in {output_path}/{name}/{version}/checkpoints and resumes it.
            Without defining a version this means no checkpoint can be found as there will not exist a  matching folder.
        """
        if version is None:
            version = _generate_version()
        if resume_checkpoint is None:
            resume_checkpoint = self._find_checkpoint(name, version, output_path)
        trainer = pl.Trainer(
            default_root_dir=output_path,
            max_epochs=getattr(self.hparams, "max_epochs", 1000),
            gpus=gpus,
            num_nodes=nodes,
            logger=TensorBoardLogger(
                save_dir=output_path, version=version, name=name
            ),
            resume_from_checkpoint=resume_checkpoint
        )
        trainer.fit(self)

    def _find_checkpoint(self, name, version, output_path):
        resume_checkpoint = None
        checkpoint_folder = os.path.join(output_path, name, version, "checkpoints")
        if os.path.exists(checkpoint_folder):
            checkpoints = sorted(os.listdir(checkpoint_folder))
            if len(checkpoints) > 0:
                resume_checkpoint = os.path.join(checkpoint_folder, checkpoints[-1])
        return resume_checkpoint

    def load_data(self, stage=None) -> Tuple[Any, Any]:
        """
        **ABSTRACT:** Load the data for training and validation.

        :return: A tuple of the train and val dataset.
        """
        raise NotImplementedError("Must be implemented by inheriting classes.")

    def training_step(self, batch, batch_idx):
        """
        Executes a training step.

        By default this calls the step function.
        :param batch: A batch of training data received from the train loader.
        :param batch_idx: The index of the batch.
        """
        feature, target = batch
        return self.step(feature, target, batch_idx)

    def validation_step(self, batch, batch_idx):
        """
        Executes a validation step.

        By default this calls the step function.
        :param batch: A batch of val data received from the val loader.
        :param batch_idx: The index of the batch.
        """
        feature, target = batch
        return self.step(feature, target, batch_idx)

    def validation_epoch_end(self, val_step_outputs):
        """
        This function is called after all training steps.

        It accumulates the loss into a val_loss which is logged in the end.
        """
        avg_loss = torch.tensor([x for x in val_step_outputs]).mean()
        return {'val_loss': avg_loss}

    def setup(self, stage=None):
        """
        This function is for setting up the training.

        The default implementation calls the load_data function and
        stores the result in self.train_data and self.val_data.
        (It is called once per process.)
        """
        self.train_data, self.val_data = self.load_data(stage=stage)

    def train_dataloader(self):
        """
        Create a training dataloader.

        The default implementation wraps self.train_data in a Dataloader.
        """
        shuffle = True
        if isinstance(self.train_data, IterableDataset):
            shuffle = False
        return DataLoader(self.train_data, batch_size=self.hparams.batch_size, shuffle=shuffle)

    def val_dataloader(self):
        """
        Create a validation dataloader.

        The default implementation wraps self.val_data in a Dataloader.
        """
        shuffle = True
        if isinstance(self.val_data, IterableDataset):
            shuffle = False
        return DataLoader(self.val_data, batch_size=self.hparams.batch_size, shuffle=shuffle)



"""doc
# pytorch_mjolnir._RemoteRunner / mjolnir_remote

> A little tool to help with running stuff on a cluster or remotely.

## Remote Run (Server or SLURM)

See the help by running.
```bash
mjolnir_remote --help
```
This will move the code from the workspace to "{user}@{host}:{remote_results_path}/{name}/{version}/src/{repository}" and then execute it.

## Run templates

The script also takes care of filling the templates with the right information to run them on the remote.

**SLURM:**
The `run.template.slurm` is a file in your cwd which must be runnable by sbatch.
It can contain `{remote_results_path}, {repository}, {partition}, {name}, {version}, {user}` which will be replaced by the arguments provided when calling `mjolnir_remote`.
Note that this script itself can call the run.sh which will be also generated in the slurm case.

**RUN:**
The `run.template.sh` is a file containing the actual code that should be run.
It can contain `{remote_results_path}, {repository}, {name}, {version}, {cmd}, {user}` which will be replaced by the arguments provided when calling `mjolnir_remote`.
"""
import os
import argparse
from pytorch_mjolnir.experiment import _generate_version


class _RemoteRunner:
    def __init__(self, host, slurm_template, user, dry_run=False, slurm=None):
        self.host = host
        self.slurm_template = slurm_template
        self.dry_run = dry_run
        self.slurm = slurm
        self.user = user

    def run(self, name, version, workspace, repository, remote_results_path, run_template, cmd):
        print(f"Preparing: {cmd}")
        def _run(cmd):
            print(f"> {cmd}")
            if not self.dry_run:
                os.system(cmd)

        slurmfile = os.path.join(workspace, repository, "run.slurm")
        if self.slurm is not None:
            with open(slurmfile, "w") as f:
                f.write(self.slurm_template.format(repository=repository, partition=self.slurm, remote_results_path=remote_results_path, name=name, version=version, user=self.user))

        runfile = os.path.join(workspace, repository, "run.sh")
        with open(runfile, "w") as f:
            f.write(run_template.format(repository=repository, remote_results_path=remote_results_path, name=name, version=version, cmd=cmd, user=self.user))

        _run(f"ssh {self.user}@{self.host} mkdir -p {remote_results_path}/{name}/{version}/src")
        _run(f"cd {workspace} && tar --exclude=.git --exclude=__pycache__ --exclude=*.egg-info --exclude=docs --exclude=.vscode -cf - . | ssh {self.user}@{self.host} tar -xf - -C {remote_results_path}/{name}/{version}/src")
        
        if os.path.exists(slurmfile):
            os.remove(slurmfile)
        if os.path.exists(runfile):
            os.remove(runfile)

        if self.slurm is not None:
            _run(f"ssh {self.user}@{self.host} sbatch {remote_results_path}/{name}/{version}/src/{repository}/run.slurm")
        else:
            _run(f"ssh {self.user}@{self.host} bash {remote_results_path}/{name}/{version}/src/{repository}/run.sh")


def main():
    slurm_host = "localhost"
    if "SLURM_HOST" in os.environ:
        slurm_host = os.environ["SLURM_HOST"]
    default_partition = "batch"
    if "SLURM_PARTITION" in os.environ:
        default_partition = os.environ["SLURM_PARTITION"]
    remote_results_path = None
    if "REMOTE_RESULTS_PATH" in os.environ:
        remote_results_path = os.environ["REMOTE_RESULTS_PATH"]
    user = os.environ["USER"]

    parser = argparse.ArgumentParser(description='The main entry point for the script.')
    parser.add_argument('--name', type=str, required=True, help='The name for the experiment.')
    parser.add_argument('--version', type=str, required=False, default=None, help='A suffix that is added to the version if you want. Otherwise it will be just the date without suffix.')
    parser.add_argument('--cmd', type=str, required=False, default="python", help='The command to run.')
    parser.add_argument('--host', type=str, required=False, default=slurm_host, help='The hostname to connect to (e.g. "server.de"), defaults to $SLURM_HOST.')
    parser.add_argument('--user', type=str, required=False, default=user, help='The username used, defaults to $USER.')
    parser.add_argument('--slurm_template', type=str, required=False, default="run.template.slurm", help='The slurm template file.')
    parser.add_argument('--run_template', type=str, required=False, default="run.template.sh", help='The run template file.')
    parser.add_argument('--repository', type=str, default=None, required=False, help='What package should be executed (defaults to cwd).')
    parser.add_argument('--workspace', type=str, default="..", required=False, help='Where the workspace with all git projects is. The entire workspace will be moved to the remote. (defaults to "..", parent directory).')
    parser.add_argument('--partition', type=str, required=False, default=default_partition, help='Select the partition on which to run the code.')
    parser.add_argument('--remote_results_path', type=str, required=(remote_results_path is None), default=remote_results_path, help='The path where on the remote the results live (defaults to $REMOTE_RESULTS_PATH).')
    parser.add_argument('--dry-run', action='store_true', help='This flag will not execute the commands on the cluster but only emulate locally.')
    args, other_args = parser.parse_known_args()

    repository = args.repository
    if repository is None:
        repository = os.getcwd().split("/")[-1]

    name = args.name
    version = _generate_version()
    if args.version is not None:
        version = f"{version}_{args.version}"

    cmd = f"{args.cmd} {' '.join(other_args)}"

    slurm = None
    if args.host == slurm_host:
        slurm = args.partition
    with open(args.run_template, "r") as f:
        run_template = f.read()

    slurm_template = ""
    if slurm:
        with open(args.slurm_template, "r") as f:
            slurm_template = f.read()

    remote_runner = _RemoteRunner(args.host, slurm_template, user, args.dry_run, slurm=slurm)
    remote_runner.run(name, version, args.workspace, repository, args.remote_results_path, run_template, cmd)


if __name__ == "__main__":
    main()



"""doc
# pytorch_mjolnir.SupervisedExperiment

> An implementation of an experiment for supervised training.

You simply set a model and a loss as attributes in the constructor and the experiment takes care of the rest.
"""

from pytorch_mjolnir.experiment import Experiment


class SupervisedExperiment(Experiment):
    """
    A supervised experiment implements forward and step by using the model and loss variable.

    In your constructor simply define:
    ```
    def __init__(self, learning_rate=1e-3, batch_size=32):
        super().__init__()
        self.save_hyperparameters()
        self.model = Model()
        self.loss = Loss()
    ```
    """
    def forward(self, *args, **kwargs):
        """
        Proxy to self.model.
        
        Arguments get passed unchanged.
        """
        return self.model(*args, **kwargs)

    def step(self, feature, target, batch_idx):
        """
        Implementation of a supervised training step.

        The output of the model will be directly given to the loss without modification.

        :param feature: A namedtuple from the dataloader that will be given to the forward as ordered parameters.
        :param target: A namedtuple from the dataloader that will be given to the loss.
        :return: The loss.
        """
        prediction = self(*feature)
        self.loss.log = self.log
        loss = self.loss(prediction, target)
        if self.training:
            self.log('loss/train', loss)
        else:
            self.log('loss/val', loss)
        return loss



from setuptools import setup, find_packages
from codecs import open
from os import path

__version__ = '0.1.0'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs if x.startswith('git+')]

setup(
    name='pytorch_mjolnir',
    version=__version__,
    description='A catalyst for experiments with lightning.',
    long_description=long_description,
    url='https://github.com/penguinmenac3/pytorch-mjolnir',
    download_url='https://github.com/penguinmenac3/pytorch-mjolnir/tarball/' + __version__,
    license='MIT',
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
    ],
    keywords='',
    packages=find_packages(exclude=['examples', 'docs', 'tests*']),
    include_package_data=True,
    author='Michael Fuerst',
    install_requires=install_requires,
    dependency_links=dependency_links,
    author_email='mail@michaelfuerst.de',
    entry_points={
        'console_scripts': [
            'mjolnir_remote = pytorch_mjolnir.remote_run:main',
        ]
    },
    extras_require={
        'dev': ['nose2', 'packaging', 'jlabdev']
    },
)
